/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

/// Touch the surface with the
/// given instrument.

#version 110


uniform sampler2D primitive;    /* Brush primitive.             */
uniform sampler2D color;        /* Brush color pool.            */
uniform sampler2D paper;        /* Paper texture.               */
uniform sampler2D oldpatch;     /* Patch to be overwritten.     */
uniform sampler2D noise_re;     /* Base noise table (real).     */
uniform sampler2D noise_im;     /* Base noise table (imag).     */
uniform vec4 ctl_mask;          /* A mask for control channel.  */
uniform float sensitivity;      /* Texture sensitivity.         */
uniform float noise_freq;       /* Perlin noise parameter.      */
uniform float noise_amp;        /* Perlin noise parameter.      */

// Painting style variables.
uniform float bleedrate;        /* Bleeding rate.               */
uniform float suckrate;         /* Sucking rate.                */
uniform float thickness;        /* Amount of deposited paint.   */

uniform vec2 patch_txmu;        /* Transform into the patch coords.     */
uniform vec2 patch_txadd;       /* Transform into the patch coords.     */
uniform vec2 color_txmu;        /* Transform into the colorfield coords.*/
uniform vec2 color_txadd;       /* Transform into the colorfield coords.*/
uniform vec2 paper_txmu;        /* Transform into the paper coords.     */
uniform vec2 paper_txadd;       /* Transform into the paper coords.     */
uniform vec2 world_txmu;        /* Transform into the canvas coords.    */
uniform vec2 world_txadd;       /* Transform into the canvas coords.    */

const vec4  g_ZERO4 = vec4 (0.0, 0.0, 0.0, 0.0);
const vec4  g_ONES4 = vec4 (1.0, 1.0, 1.0, 1.0);
const float g_2PI   = 6.28318531;



vec4 perlin_noise4 (vec2 tx) {
    // Base noise look up.
    vec4 hi_r = texture2D (noise_re, tx);
    vec4 hi_i = texture2D (noise_im, tx);
    vec4 lo_r = texture2D (noise_re, tx * 0.125);
  //vec4 lo_i = texture2D (noise_im, tx * 0.125);

    // Compose a single noise value from
    // the low- and high-frequency components.
  //vec4 rotated_r = hi_r * cos (g_2PI * lo_r);
  //vec4 rotated_i = hi_i * sin (g_2PI * lo_r);
    vec4 cos2  = cos (g_2PI * lo_r);
         cos2 *= cos2;

    vec4 rotated_r2 = hi_r * hi_r * cos2;
    vec4 rotated_i2 = hi_i * hi_i * (g_ONES4 - cos2);

    return sqrt (rotated_r2 + rotated_i2);
}

/**
 * for (int i = 0; i < octaves; i++) {
 *     total += perlin_noise4 (tx * frequency) * amplitude;
 *     frequency *= 2.0;
 *     amplitude *= persistence;
 * }
 *
 * total = clamp (total * cloud_mu - cloud_sub, 0.0, 1.0);
 */
vec4 perlin_noise_composite4 (vec2 tx, float frequency, float amplitude) {
    return clamp (
        (perlin_noise4 (tx *       frequency) * amplitude +
         perlin_noise4 (tx * 2.0 * frequency) * amplitude +
         perlin_noise4 (tx * 4.0 * frequency) * amplitude) * 2.0 +
         vec4 (-0.5, -0.5, -0.5, -0.5),
        0.0,
        1.0
    );
}



/// Here you can place your custom
/// shader for an instrument.
vec4 custom (vec4 newcolor, vec4 oldcolor, float ctl) {
    return newcolor;
}

/// gl_TexCoord[0] -- position in mask.
/// gl_TexCoord[1] -- position in patch.
void main (void) {

    // Primitive as mask.
    float mask = texture2D (primitive, gl_TexCoord[0].st).x;
    if (mask < 0.5)
        discard;

    // Merge to reduce the number of multiplications.
  //vec2 tx_patch = gl_TexCoord[1].st * patch_txmu + patch_txadd;
  //vec2 tx_color = gl_TexCoord[1].st * color_txmu + color_txadd;
    vec4 tx0 = vec4 (gl_TexCoord[1].st, gl_TexCoord[1].st);
    vec4 tx1 = tx0;

    tx0 = tx0 * vec4 (patch_txmu,  color_txmu) +
                vec4 (patch_txadd, color_txadd);
    tx1 = tx1 * vec4 (paper_txmu,  world_txmu) +
                vec4 (paper_txadd, world_txadd);

    vec4 brush = texture2D (color,     tx0.ba);
    vec4 patch = texture2D (oldpatch,  tx0.rg);

    float ctl_src = dot (brush, ctl_mask);
    float ctl_dst = dot (patch, abs (ctl_mask));
    float height  = texture2D (paper, fract (tx1.rg)).x;

    //ctl_src -= 1.0 - step (0.01, ctl_src);
    //if (ctl_src < 0.01)
        //ctl_src = -1.0;

    // Flow direction.
    float ctl = clamp (ctl_src + 0.1 - 0.5 * (1.0 - bleedrate) * ctl_dst, -1.0, +1.0);

    // Rough sensitivity.
    float penetration = sensitivity;
    //if (penetration >= height)
    //    discard;

    float penetration1 = (height - penetration) / height;
  //float penetration1 = smoothstep (height, penetration);

    // Scale (fine sensitivity?).
    float ctl1 = ctl;
    ctl *= penetration1 * mask * mask;

    // Flow control.
    if (ctl1 <= 0.0) {
        ctl = max (0.1, -ctl);

        // Instrument is sucking in.
        vec4 faded   = mix (patch, g_ZERO4, ctl * bleedrate);
        gl_FragColor = max (faded, vec4 (brush.rgb, 0.0));
    } else {
      //vec2 world_tx = tx_patch * world_txmu + world_txadd;
        vec4 noise = perlin_noise_composite4 (
                         tx1.ba,
                         noise_freq,
                         noise_amp
                     );

        // * step (penetration, height)
        ctl = max (0.02, ctl - noise.r);
        vec4 color = brush;

        float w0 = max (0.02, patch.a);
        float w1 = ctl * bleedrate * 0.5;
        vec4 mixed = (patch * w0 + color * w1) / (w0 + w1);

        mixed.a = mix (patch.a, color.a, ctl * thickness);

        // Instrument is bleeding out.
        gl_FragColor = custom (mixed, patch, ctl);
    }
}
