/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

// Kubelka-Munk layer rendering.
//

#version 110


/* Absorbtion and scattering of a new layer;      */
/* reflectance and transmittance of a last layer. */
uniform sampler2D KSRT_center;
uniform sampler2D KSRT_left;
uniform sampler2D KSRT_right;
uniform sampler2D KSRT_top;
uniform sampler2D KSRT_bottom;

uniform sampler2D Meta0;   /* Patch metadata.                       */

uniform vec4 J;            /* Source (upward) ray.                  */
uniform vec4 I;            /* Light (downward) ray.                 */
uniform float opacity;     /* Layer opacity.                        */

/* Bump mapping parameters. */
uniform sampler2D paper_bump;
uniform vec2 bump_off;
uniform vec2 bump_scale;
uniform float paper_embossing;
uniform float paint_embossing;

// A gift to us from a vertex program. ^^'
varying vec4 lightDir;

const vec4 g_ZERO = vec4 (0.00, 0.00, 0.00, 0.00);
const vec4 g_ONE  = vec4 (1.00, 1.00, 1.00, 1.00);
const vec4 g_T    = vec4 (0.01, 0.01, 0.01, 0.00);

const vec4 g_BUMPMUL = vec4 (+2.00, +2.00, +2.00, +1.00);
const vec4 g_BUMPADD = vec4 (-1.00, -1.00, -1.00, +0.00);
const vec4 g_THR     = vec4 (+0.50, +0.00, +0.00, +0.50);
const vec4 g_STEP    = vec4 (0.002, 0.000, 0.000, 0.002);
const vec4 g_SFNORM  = vec4 (0.0, 0.0, 1.0, 0.0);

void kmimport (vec4 k, vec4 s, out vec4 r, out vec4 t, float z) {
    /**
     * Translate the parameters.
     * Let: sinh(x) ~= x;
     *      cosh(x) ~= 0.5 * x * x + 1;
     *      sqrt (x * x - 1) ~= x;
     */
    vec4 alpha = (k + s) / s;                               //
    vec4 beta  = sqrt (alpha * alpha - g_ONE);              //
  //vec4 beta  = alpha;                                     // [approximated]
    vec4 theta = z * s * beta;                              //
    vec4 epp   = exp (+theta);                              //
    vec4 epm   = exp (-theta);                              //
  //vec4 sih   = theta;                                     // [approximated]
    vec4 sih   = 0.5 * (epp - epm);                         //
  //vec4 coh   = g_ONE + 0.5 * theta * theta;               // [approximated]
    vec4 coh   = 0.5 * (epp + epm);                         //
    vec4 denum = g_ONE / (alpha * sih + beta * coh);        //
    vec4 r1    = sih  * denum;                              // layer reflectance.
    vec4 t1    = beta * denum;                              // layer transmittance.
                                                            //
    //r1    = s * z;                                        // layer reflectance.
    //t1    = g_ONE - k * z;                                // layer transmittance.

    r = r1;
    t = t1;
}

void main (void) {
    vec2 tx0 = gl_TexCoord[0].st;                           // Left-top sub-patch.
    vec2 tx1 = tx0 + vec2 (0.5, 0.0);                       // Right-top sub-patch.
    vec2 tx2 = tx0 + vec2 (0.0, 0.5);                       // Left-bottom sub-patch.
    vec2 tx3 = tx0 + vec2 (0.5, 0.5);                       // Right-bottom sub-patch.
    vec4  k1 = max (g_T, texture2D (KSRT_center, tx0));
    vec4  s1 = max (g_T, texture2D (KSRT_center, tx1));
    vec4  r0 = texture2D (KSRT_center, tx2);
    vec4  t0 = texture2D (KSRT_center, tx3);

    // Texture was filled with (1 - R) values.
    r0 = g_ONE - r0;

    vec4 r1, r2;
    vec4 t1, t2;

    // Each patch contains two layers.
    kmimport (2.0 * k1, 20.0 * s1, r1, t1, k1.a);
  //kmimport (2.0 * k2, 20.0 * s2, r2, t2, k2.a);

    // Merge the layers:
    // Wet channel.
    vec4 r02 = r1;
    vec4 t02 = t1;

    // Dry channel (doesn't exist anymore).
  //vec4 denum = g_ONE / (g_ONE - r01 * r2);
  //vec4 r02 = r01 + t01 * t01 * r2 * denum;     // layer reflectance.
  //vec4 r02 = r01 + t2 * t2 * r2 * denum;       // ???
  //vec4 t02 = t01 * t2 * denum;                 // layer transmittance.

    // Scene background.
    vec4 denum = g_ONE / (g_ONE - r02 * r0);
    vec4 r03 = r02 + t02 * t02 * r0 * denum;     // overall reflectance.
    vec4 t03 = t02 * t0 * denum;                 // overall transmittance.

    //r03 = vec4 (1.0, 0.0, 0.0, 0.0);
    //t03 = vec4 (1.0, 1.0, 0.0, 0.0);

    vec4 tx_lt0 = vec4 (tx0, tx0) - g_STEP;
    vec4 tx_rb0 = vec4 (tx0, tx0) + g_STEP;
    vec4 tx_lt1 = tx_lt0 + g_THR;
    vec4 tx_rb1 = tx_rb0 - g_THR;

    // Paint bump map.
    // Evaluate the neighbour patches.
    float left   = mix (
                     texture2D (KSRT_center, tx_lt0.xy),
                     texture2D (KSRT_left,   tx_lt1.xy),
                     step (tx_lt0.x, 0.0)
                   ).a;
    float top    = mix (
                     texture2D (KSRT_center, tx_lt0.zw),
                     texture2D (KSRT_top,    tx_lt1.zw),
                     step (tx_lt0.w, 0.0)
                   ).a;
    float right  = mix (
                     texture2D (KSRT_center, tx_rb0.xy),
                     texture2D (KSRT_right,  tx_rb1.xy),
                     step (0.5, tx_rb0.x)
                   ).a;
    float bottom = mix (
                     texture2D (KSRT_center, tx_rb0.zw),
                     texture2D (KSRT_bottom, tx_rb1.zw),
                     step (0.5, tx_rb0.w)
                   ).a;

    vec4 d = vec4 (right, bottom, 1.0, 0.0) -
             vec4 (left,  top,    0.0, 0.0);

    // Bump mapping.
    vec2 bump_tx = fract (tx0 * bump_scale + bump_off);             // Paper texcoords.
    vec4 bump0 = texture2D (paper_bump, bump_tx);                   // Paper bump.
    vec4 bump1 = texture2D (Meta0, tx1);                            // Dry bump.
         bump1 = g_BUMPMUL * bump1 + g_BUMPADD;                     // Denormalize.
    vec4 bump2 = normalize (d);                                     // Wet bump.

    float n = (k1.a * paint_embossing + bump1.a);
    float w = step (0.01, paint_embossing);

    // Compose a resulting bump vector.
    bump1 = (bump1 * bump1.a + bump2 * k1.a * paint_embossing) /
            (n + 0.001);                                            // Combined bump.
    vec4 bump  = mix (g_SFNORM, bump0, paper_embossing);            // Resulting paper bump.
         bump  = mix (bump, bump1, min (1.0, 10.0 * w * n));

    // Normal-to-light.
    float mu = dot (bump, lightDir);

    // KM-glazing.
    vec4 j1 = clamp (mu * r03, 0.0, 1.0);

    // FIXME Blinn-Phong highlights?

    gl_FragData[0] = vec4 (j1.rgb, 1.0);
}
