/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

// Paints drying iteration.
//

#version 110


uniform sampler2D optical;    /** Layer absorbing and scattering,
                                * Background reflectance and transmittance. */

uniform sampler2D paper;
uniform vec2      paper_off;
uniform vec2      paper_scale;
uniform float     dt;
uniform float     granulation;
uniform float     ballance1;
uniform float     ballance2;

const vec2 tx_dx = vec2 (1.0 / 512.0, 0.0);
const vec2 tx_dy = vec2 (0.0, 1.0 / 512.0);

const vec4 g_STEP = vec4 (1.000, 1.000, 1.000, 0.001);
const vec4 g_ZERO = vec4 (0.000, 0.000, 0.000, 0.000);
const vec4 g_ONE  = vec4 (1.000, 1.000, 1.000, 1.000);
const vec4 g_T    = vec4 (0.01, 0.01, 0.01, 0.00);

void kmimport (vec4 k, vec4 s, out vec4 r, out vec4 t, float z) {
    /**
     * Translate the parameters.
     * Let: sinh(x) ~= x;
     *      cosh(x) ~= 0.5 * x * x + 1;
     *      sqrt (x * x - 1) ~= x;
     */
    vec4 alpha = (k + s) / s;
    vec4 beta  = sqrt (alpha * alpha - g_ONE);
  //vec4 beta  = alpha;                                     // [approximated]
    vec4 theta = z * s * beta;                              //
    vec4 epp   = exp (+theta);
    vec4 epm   = exp (-theta);
  //vec4 sih   = theta;                                     // [approximated]
    vec4 sih   = 0.5 * (epp - epm);
  //vec4 coh   = g_ONE + 0.5 * theta * theta;               // [approximated]
    vec4 coh   = 0.5 * (epp + epm);
    vec4 denum = g_ONE / (alpha * sih + beta * coh);        //
    vec4 r1    = sih  * denum;                              // layer reflectance.
    vec4 t1    = beta * denum;                              // layer transmittance.

    //r1    = s * z;                                        // layer reflectance.
    //t1    = g_ONE - k * z;                                // layer transmittance.

    r = r1;
    t = t1;
}


void main (void) {
    vec2 tx  = gl_TexCoord[0].st;
    vec2 tx0 = 0.5 * tx;
    vec2 tx1 = tx0 + vec2 (0.5, 0.0);
    vec2 tx2 = tx0 + vec2 (0.0, 0.5);
    vec2 tx3 = tx0 + vec2 (0.5, 0.5);

    vec4 r1     = texture2D (optical, tx2);
    vec4 t1     = texture2D (optical, tx3);
    vec4 src_a  = max (g_T, texture2D (optical, tx0));
    vec4 src_s  = max (g_T, texture2D (optical, tx1));

    // Texture was filled with (1 - R) values.
    r1 = g_ONE - r1;

    vec4 h = texture2D (paper, fract (gl_TexCoord[1].st * paper_scale + paper_off));

    vec4 rate = vec4 (0.0, 0.0, 0.0,
                      1.0 - granulation * (1.0 - h.x));

    // Flow rate and constrains.
    rate *= dt;
    vec4 flow_a  = min (rate, src_a.a);
    vec4 flow_s  = min (rate, src_s.a);

    // Use Sobel operator to sharpen the edges.
    float edge = 0.0;
                //abs (dFdx (src_a.a)) + abs (dFdy (src_a.a));
/*
    length (vec2 (
        texture2D (wet_a, tx - tx_dx - tx_dy).a +
        texture2D (wet_a, tx -         tx_dy).a +
        texture2D (wet_a, tx + tx_dx - tx_dy).a -
        texture2D (wet_a, tx - tx_dx + tx_dy).a -
        texture2D (wet_a, tx +         tx_dy).a -
        texture2D (wet_a, tx + tx_dx + tx_dy).a,
        texture2D (wet_a, tx - tx_dx + tx_dy).a +
        texture2D (wet_a, tx - tx_dx        ).a +
        texture2D (wet_a, tx - tx_dx - tx_dy).a -
        texture2D (wet_a, tx + tx_dx - tx_dy).a -
        texture2D (wet_a, tx + tx_dx        ).a -
        texture2D (wet_a, tx + tx_dx + tx_dy).a
    ));
*/

    vec4 r0;
    vec4 t0;
    kmimport (2.0 * src_a, 20.0 * src_s, r0, t0, flow_a.a);

    // Merge the layers.
    vec4 denum = g_ONE / (g_ONE - r0 * r1);
    vec4 r = r0 + t0 * t0 * r1 * denum;             // overall reflectance.
    vec4 t = t0 * t1 * denum;                       // overall transmittance.

    // Constrain.
    vec4 a1 = src_a - flow_a;
    vec4 s1 = src_s - flow_s;

    a1 = a1 * step (0.01, a1);
    s1 = s1 * step (0.01, s1);
    r  = g_ONE - r;

    gl_FragData[0] = mix (
                         mix (a1, s1, ballance1),
                         mix (r,  t,  ballance1),
                         ballance2
                     );

/*
 * Uncomment to disable the procedure.
 *
    gl_FragData[0] = mix (
                         mix (src_a, src_s, ballance1),
                         mix (g_ONE - r1, t1, ballance1),
                         ballance2
                     );
*/
}
