/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

// Paints drying iteration:
// paints bump map propagation.

#version 110

uniform sampler2D source;
uniform sampler2D optical_center;
uniform sampler2D optical_left;
uniform sampler2D optical_right;
uniform sampler2D optical_top;
uniform sampler2D optical_bottom;
uniform float     dt;
uniform float     embossing;

const vec4 g_T    = vec4 (+0.01, +0.01, +0.01, +0.00);
const vec4 g_MUL0 = vec4 (+2.00, +2.00, +2.00, +1.00);
const vec4 g_MUL1 = vec4 (+0.50, +0.50, +0.50, +1.00);
const vec4 g_ADD0 = vec4 (-1.00, -1.00, -1.00, +0.00);
const vec4 g_ADD1 = vec4 (+0.50, +0.50, +0.50, +0.00);
const vec4 g_THR  = vec4 (+0.50, +0.00, +0.00, +0.50);
const vec4 g_STEP = vec4 (0.002, 0.000, 0.000, 0.002);


void main (void) {
    vec2 tx0 = 0.5 * gl_TexCoord[0].st;                     // Left-top sub-patch.
    vec2 tx1 = tx0 + vec2 (0.5, 0.0);                       // Right-top sub-patch.
    vec4 bump0 = texture2D (source,  tx1);                  // Old bump.
         bump0 = g_MUL0 * bump0 + g_ADD0;                   // Denormalize.

    vec4 k1 = max (g_T, texture2D (optical_center, tx0));
    vec4 s1 = max (g_T, texture2D (optical_center, tx1));

    vec4 tx_lt0 = vec4 (tx0, tx0) - g_STEP;
    vec4 tx_rb0 = vec4 (tx0, tx0) + g_STEP;
    vec4 tx_lt1 = tx_lt0 + g_THR;
    vec4 tx_rb1 = tx_rb0 - g_THR;

    // Paint bump map.
    // Evaluate the neighbour patches.
    float left   = mix (
                     texture2D (optical_center, tx_lt0.xy),
                     texture2D (optical_left,   tx_lt1.xy),
                     step (tx_lt0.x, 0.0)
                   ).a;
    float top    = mix (
                     texture2D (optical_center, tx_lt0.zw),
                     texture2D (optical_top,    tx_lt1.zw),
                     step (tx_lt0.w, 0.0)
                   ).a;
    float right  = mix (
                     texture2D (optical_center, tx_rb0.xy),
                     texture2D (optical_right,  tx_rb1.xy),
                     step (0.5, tx_rb0.x)
                   ).a;
    float bottom = mix (
                     texture2D (optical_center, tx_rb0.zw),
                     texture2D (optical_bottom, tx_rb1.zw),
                     step (0.5, tx_rb0.w)
                   ).a;

    vec4 d = vec4 (right, bottom, 1.0, 0.0) -
             vec4 (left,  top,    0.0, 0.0);

    if (embossing < 0.01)
        discard;

    // Bump update.
    float dz = dt * step (0.001, k1.a * embossing);
    float a  = bump0.a;
    vec4 bump1   = normalize (d);
         bump1   = (bump1 * dz + bump0 * a) / (a + dz);     // Melt in.
         bump1   = g_MUL1 * bump1 + g_ADD1;                 // Normalize.
         bump1.a = min (1.0, bump0.a + embossing * dz);     //

    gl_FragData[0] = bump1;
}
