/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

/// Touch the surface with the
/// given instrument.

#version 140

uniform sampler2D control_src;  /* Flow control plane.          */
uniform sampler2D control_dst;  /* Flow control plane.          */
uniform sampler2D primitive;    /* Brush primitive.             */
uniform sampler2D paper;        /* Paper texture.               */
uniform sampler2D oldpatch;     /* Patch to be overwritten.     */
uniform vec4 ctl_mask;          /* A mask for control channel.  */
uniform float sensitivity;      /* Texture sensitivity.         */
uniform float bleedrate;        /* Bleeding rate.               */
uniform float suckrate;         /* Sucking rate.                */

// Painting style variables.
uniform vec4 ps_channel_mu;     /* Per-component multipliers.   */

uniform vec2 paper_offset;      /* Offset of a paper texture.           */
uniform vec2 paper_scale;       /* Scaling factor for a paper texture.  */

const vec3 g_ZERO3 = vec3 (0.0, 0.0, 0.0);

centroid noperspective in vec2 g_texcoord0;
centroid noperspective in vec2 g_texcoord1;

out vec4 result;


/// Here you can place your custom
/// shader for an instrument.
vec3 custom (vec3 newcolor, vec3 oldcolor, float control) {
    return clamp (newcolor * control + oldcolor, oldcolor, newcolor);
}

void main (void) {
    float ctl_src = dot (texture (control_src, g_texcoord0), ctl_mask);
    float ctl_dst = dot (texture (control_dst, g_texcoord1), ctl_mask);

    vec4 brush   = texture (primitive, g_texcoord0);
    vec4 patch   = texture (oldpatch,  g_texcoord0);
    float height = texture (paper, fract (g_texcoord0 * paper_scale + paper_offset)).x;

    float ctl = clamp (bleedrate * (ctl_src - ctl_dst), -1.0, +1.0);

    // Source control primitive as mask.
    if (abs (ctl_src) < 0.5)
        discard;

    // Fine sensitivity.
    float penetration = sensitivity * 0.9;
    if (penetration > height)
        discard;

    // Scale.
    if (penetration > 0.1f)
        ctl *= (height - penetration) / penetration;

    // Flow control.
    if (ctl <= 0.0) {
        // Instrument is sucking in.
        result = vec4 (mix (patch.rgb, g_ZERO3, 0.0 - ctl), 1.0);
    } else {
        // Instrument is bleeding out.
        result = ps_channel_mu * vec4 (custom (brush.rgb, patch.rgb, ctl), 1.0);
    }
}
