/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */

// Kubelka-Munk layer rendering.

#version 140

uniform sampler2D KSRT;      /* Absorbtion and scattering of a new layer;      */
                             /* reflectance and transmittance of a last layer. */
uniform sampler2D Meta0;     /* Patch metadata.                                */

uniform vec4 J;            /* Source (upward) ray.                  */
uniform vec4 I;            /* Light (downward) ray.                 */
uniform float opacity;     /* Layer opacity.                        */

/* Bump mapping parameters. */
uniform sampler2D paper_bump;
uniform vec2 bump_off;
uniform vec2 bump_scale;
uniform float embossing;

const vec4 g_ZERO = vec4 (0.00, 0.00, 0.00, 0.00);
const vec4 g_ONE  = vec4 (1.00, 1.00, 1.00, 1.00);
const vec4 g_T    = vec4 (0.01, 0.01, 0.01, 0.00);

// A gift to us from a vertex program. ^^'
centroid nonperspective vec4 lightDir;
centroid nonperspective out vec2 g_texcoord0;
centroid nonperspective out vec2 g_texcoord1;

out vec4 result;


void kmimport (in vec4 k, in vec4 s, out vec4 r, out vec4 t, in float z) {
    /**
     * Translate the parameters.
     * Let: sinh(x) ~= x;
     *      cosh(x) ~= 0.5 * x * x + 1;
     *      sqrt (x * x - 1) ~= x;
     */
    lowp vec4 alpha = (k + s) / s;                               //
    lowp vec4 beta  = sqrt (alpha * alpha - g_ONE);              //
  //lowp vec4 beta  = alpha;                                     // [approximated]
    lowp vec4 theta = z * s * beta;                              //
    lowp vec4 epp   = exp (+theta);                              //
    lowp vec4 epm   = exp (-theta);                              //
  //lowp vec4 sih   = theta;                                     // [approximated]
    lowp vec4 sih   = 0.5 * (epp - epm);                         //
  //lowp vec4 coh   = g_ONE + 0.5 * theta * theta;               // [approximated]
    lowp vec4 coh   = 0.5 * (epp + epm);                         //
    lowp vec4 denum = g_ONE / (alpha * sih + beta * coh);        //
    lowp vec4 r1    = sih  * denum;                              // layer reflectance.
    lowp vec4 t1    = beta * denum;                              // layer transmittance.
                                                            //
    //r1    = s * z;                                        // layer reflectance.
    //t1    = g_ONE - k * z;                                // layer transmittance.

    r = r1;
    t = t1;
}

void main (void) {
    vec2 tx0 = g_texcoord0;                                 // Left-top sub-patch.
    vec2 tx1 = g_texcoord0 + vec2 (0.5, 0.0);               // Right-top sub-patch.
    vec2 tx2 = g_texcoord0 + vec2 (0.0, 0.5);               // Left-bottom sub-patch.
    vec2 tx3 = g_texcoord0 + vec2 (0.5, 0.5);               // Right-bottom sub-patch.
    vec4  k1 = max (g_T, texture2D (KSRT, tx0));
    vec4  s1 = max (g_T, texture2D (KSRT, tx1));
    vec4  r0 = texture2D (KSRT, tx2);
    vec4  t0 = texture2D (KSRT, tx3);

    // Texture was filled with (1 - R) values.
    r0 = g_ONE - r0;

    vec4 r1, r2;
    vec4 t1, t2;

    // Each patch contains two layers.
    kmimport (2.0 * k1, 20.0 * s1, r1, t1, k1.a);
  //kmimport (2.0 * k2, 20.0 * s2, r2, t2, k2.a);

    // Merge the layers:
    // Wet channel.
    vec4 r02 = r1;
    vec4 t02 = t1;

    // Dry channel (doesn't exist anymore).
  //vec4 denum = g_ONE / (g_ONE - r01 * r2);
  //vec4 r02 = r01 + t01 * t01 * r2 * denum;     // layer reflectance.
  //vec4 r02 = r01 + t2 * t2 * r2 * denum;       // ???
  //vec4 t02 = t01 * t2 * denum;                 // layer transmittance.

    // Scene background.
    vec4 denum = g_ONE / (g_ONE - r02 * r0);
    vec4 r03 = r02 + t02 * t02 * r0 * denum;     // overall reflectance.
    vec4 t03 = t02 * t0 * denum;                 // overall transmittance.

    //r03 = vec4 (1.0, 0.0, 0.0, 0.0);
    //t03 = vec4 (1.0, 1.0, 0.0, 0.0);

    float left   = texture2D (KSRT, tx0 - vec2 (0.002, 0.000)).a;
    float top    = texture2D (KSRT, tx0 - vec2 (0.000, 0.002)).a;
    float right  = texture2D (KSRT, tx0 + vec2 (0.002, 0.000)).a;
    float bottom = texture2D (KSRT, tx0 + vec2 (0.000, 0.002)).a;

    vec4 d = vec4 (right, bottom, 1.0, 0.0) -
             vec4 (left,  top,    0.0, 0.0);

    vec4 paint_bump = normalize (d) +           // Wet bump.
                      texture2D (Meta0, tx1);   // Dry bump.

    // Bump mapping.
    vec4 bump0 = paint_bump;
    vec4 bump1 = texture2D (paper_bump, fract (g_texcoord0 * bump_scale + bump_off));
    vec4 bump  = embossing * bump0 + bump1;

    // Normal-to-light.
    float mu = dot (bump1, lightDir);

    // KM-glazing.
    vec4 j1 = clamp (mu * r03, 0.0, 1.0);

    // FIXME Blinn-Phong highlights?

    result = vec4 (j1.rgb, 1.0);
}
